//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <cxxtls/utf.h>
#include <portable_io.h>

namespace cxxtls
{

namespace utf
{

char const *encodingName(encoding e)
{
    switch(e)
    {
        default:  return "invalid encoding type";

        case utf8_enc:     return "utf8:    8  bit UNICODE";
        case utf16be_enc:  return "utf16be: 16 bit, big endian UNICODE";
        case utf16le_enc:  return "utf16le: 16 bit, little endian UNICODE";
        case utf32be_enc:  return "utf32be: 32 bit, big endian UNICODE";
        case utf32le_enc:  return "utf32le: 32 bit, little endian UNICODE";
        case utfunk_enc:   return "utfunk:  unknown, bad Byte Order Mark"; 
        case utf7_enc:     return "utf7:    ANSI standard text";
        case utfbin_enc:   return "utfbin:  non-text binary";

    }

    return "invalid encoding type";
}

unsigned char conflate(wchar_t c)
{
   unsigned char ch('~');


   // see http://www.utf8-chartable.de/unicode-utf8-table.pl (... ?start=256, for example)

   switch(c)
   {
      default:  ch = (c&0xff);  break;

      case 0xA1:    ch = '!'; break; // INVERTED EXCLAMATION MARK
      case 0xA2:    ch = 'c'; break; // CENT SIGN
      case 0xA3:    ch = 'L'; break; // POUND SIGN
      case 0xA4:    ch = '$'; break; // CURRENCY SIGN
      case 0xA5:    ch = 'Y'; break; // YEN SIGN
      case 0xA6:    ch = '|'; break; // BROKEN BAR
      case 0xA7:    ch = 'S'; break; // SECTION SIGN
      case 0xA8:    ch = '~'; break; // DIAERESIS
      case 0xA9:    ch = '@'; break; // COPYRIGHT SIGN
      case 0xAA:    ch = 'V'; break; // FEMININE ORDINAL INDICATOR
      case 0xAB:    ch = '<'; break; // LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
      case 0xAC:    ch = '^'; break; // NOT SIGN
      case 0xAD:    ch = '-'; break; // SOFT HYPHEN
      case 0xAE:    ch = '@'; break; // REGISTERED SIGN
      case 0xAF:    ch = '-'; break; // MACRON
      case 0xB0:    ch = 'd'; break; // DEGREE SIGN
      case 0xB1:    ch = '+'; break; // PLUS-MINUS SIGN
      case 0xB2:    ch = '2'; break; // SUPERSCRIPT TWO
      case 0xB3:    ch = '3'; break; // SUPERSCRIPT THREE
      case 0xB4:    ch = '`'; break; // ACUTE ACCENT
      case 0xB5:    ch = 'u'; break; // MICRO SIGN
      case 0xB6:    ch = 'P'; break; // PILCROW SIGN
      case 0xB7:    ch = '.'; break; // MIDDLE DOT
      case 0xB8:    ch = ','; break; // CEDILLA
      case 0xB9:    ch = '1'; break; // SUPERSCRIPT ONE
      case 0xBA:    ch = '^'; break; // MASCULINE ORDINAL INDICATOR
      case 0xBB:    ch = '>'; break; // RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
      case 0xBC:    ch = '4'; break; // VULGAR FRACTION ONE QUARTER
      case 0xBD:    ch = '2'; break; // VULGAR FRACTION ONE HALF
      case 0xBE:    ch = '3'; break; // VULGAR FRACTION THREE QUARTERS
      case 0xBF:    ch = '?'; break; // INVERTED QUESTION MARK
      case 0xC0:    ch = 'A'; break; // LATIN CAPITAL LETTER A WITH GRAVE
      case 0xC1:    ch = 'a'; break; // LATIN CAPITAL LETTER A WITH ACUTE
      case 0xC2:    ch = 'a'; break; // LATIN CAPITAL LETTER A WITH CIRCUMFLEX
      case 0xC3:    ch = 'a'; break; // LATIN CAPITAL LETTER A WITH TILDE
      case 0xC4:    ch = 'a'; break; // LATIN CAPITAL LETTER A WITH DIAERESIS
      case 0xC5:    ch = 'a'; break; // LATIN CAPITAL LETTER A WITH RING ABOVE
      case 0xC6:    ch = 'e'; break; // LATIN CAPITAL LETTER AE
      case 0xC7:    ch = 'C'; break; // LATIN CAPITAL LETTER C WITH CEDILLA
      case 0xC8:    ch = 'e'; break; // LATIN CAPITAL LETTER E WITH GRAVE
      case 0xC9:    ch = 'e'; break; // LATIN CAPITAL LETTER E WITH ACUTE
      case 0xCA:    ch = 'e'; break; // LATIN CAPITAL LETTER E WITH CIRCUMFLEX
      case 0xCB:    ch = 'e'; break; // LATIN CAPITAL LETTER E WITH DIAERESIS
      case 0xCC:    ch = 'I'; break; // LATIN CAPITAL LETTER I WITH GRAVE
      case 0xCD:    ch = 'i'; break; // LATIN CAPITAL LETTER I WITH ACUTE
      case 0xCE:    ch = 'i'; break; // LATIN CAPITAL LETTER I WITH CIRCUMFLEX
      case 0xCF:    ch = 'i'; break; // LATIN CAPITAL LETTER I WITH DIAERESIS
      case 0xD0:    ch = 'E'; break; // LATIN CAPITAL LETTER ETH
      case 0xD1:    ch = 'N'; break; // LATIN CAPITAL LETTER N WITH TILDE
      case 0xD2:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH GRAVE
      case 0xD3:    ch = 'o'; break; // LATIN CAPITAL LETTER O WITH ACUTE
      case 0xD4:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH CIRCUMFLEX
      case 0xD5:    ch = 'o'; break; // LATIN CAPITAL LETTER O WITH TILDE
      case 0xD6:    ch = 'o'; break; // LATIN CAPITAL LETTER O WITH DIAERESIS
      case 0xD7:    ch = '*'; break; // MULTIPLICATION SIGN
      case 0xD8:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH STROKE
      case 0xD9:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH GRAVE
      case 0xDA:    ch = 'u'; break; // LATIN CAPITAL LETTER U WITH ACUTE
      case 0xDB:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH CIRCUMFLEX
      case 0xDC:    ch = 'u'; break; // LATIN CAPITAL LETTER U WITH DIAERESIS
      case 0xDD:    ch = 'U'; break; // LATIN CAPITAL LETTER Y WITH ACUTE
      case 0xDE:    ch = 'T'; break; // LATIN CAPITAL LETTER THORN
      case 0xDF:    ch = 'S'; break; // LATIN SMALL LETTER SHARP S
      case 0xE0:    ch = 'A'; break; // LATIN SMALL LETTER A WITH GRAVE
      case 0xE1:    ch = 'a'; break; // LATIN SMALL LETTER A WITH ACUTE
      case 0xE2:    ch = 'A'; break; // LATIN SMALL LETTER A WITH CIRCUMFLEX
      case 0xE3:    ch = 'a'; break; // LATIN SMALL LETTER A WITH TILDE
      case 0xE4:    ch = 'A'; break; // LATIN SMALL LETTER A WITH DIAERESIS
      case 0xE5:    ch = 'a'; break; // LATIN SMALL LETTER A WITH RING ABOVE
      case 0xE6:    ch = 'e'; break; // LATIN SMALL LETTER AE
      case 0xE7:    ch = 'C'; break; // LATIN SMALL LETTER C WITH CEDILLA
      case 0xE8:    ch = 'E'; break; // LATIN SMALL LETTER E WITH GRAVE
      case 0xE9:    ch = 'e'; break; // LATIN SMALL LETTER E WITH ACUTE
      case 0xEA:    ch = 'E'; break; // LATIN SMALL LETTER E WITH CIRCUMFLEX
      case 0xEB:    ch = 'E'; break; // LATIN SMALL LETTER E WITH DIAERESIS
      case 0xEC:    ch = 'I'; break; // LATIN SMALL LETTER I WITH GRAVE
      case 0xED:    ch = 'i'; break; // LATIN SMALL LETTER I WITH ACUTE
      case 0xEE:    ch = 'I'; break; // LATIN SMALL LETTER I WITH CIRCUMFLEX
      case 0xEF:    ch = 'i'; break; // LATIN SMALL LETTER I WITH DIAERESIS
      case 0xF0:    ch = 'E'; break; // LATIN SMALL LETTER ETH
      case 0xF1:    ch = 'N'; break; // LATIN SMALL LETTER N WITH TILDE
      case 0xF2:    ch = 'O'; break; // LATIN SMALL LETTER O WITH GRAVE
      case 0xF3:    ch = 'o'; break; // LATIN SMALL LETTER O WITH ACUTE
      case 0xF4:    ch = 'O'; break; // LATIN SMALL LETTER O WITH CIRCUMFLEX
      case 0xF5:    ch = 'o'; break; // LATIN SMALL LETTER O WITH TILDE
      case 0xF6:    ch = 'O'; break; // LATIN SMALL LETTER O WITH DIAERESIS
      case 0xF7:    ch = '/'; break; // DIVISION SIGN
      case 0xF8:    ch = 'O'; break; // LATIN SMALL LETTER O WITH STROKE
      case 0xF9:    ch = 'U'; break; // LATIN SMALL LETTER U WITH GRAVE
      case 0xFA:    ch = 'u'; break; // LATIN SMALL LETTER U WITH ACUTE
      case 0xFB:    ch = 'U'; break; // LATIN SMALL LETTER U WITH CIRCUMFLEX
      case 0xFC:    ch = 'U'; break; // LATIN SMALL LETTER U WITH DIAERESIS
      case 0xFD:    ch = 'Y'; break; // LATIN SMALL LETTER Y WITH ACUTE
      case 0xFE:    ch = 't'; break; // LATIN SMALL LETTER THORN
      case 0xFF:    ch = 'y'; break; // LATIN SMALL LETTER Y WITH DIAERESIS

      case 0x100:    ch = 'A'; break; // LATIN CAPITAL LETTER A WITH MACRON
      case 0x101:    ch = 'a'; break; // LATIN SMALL LETTER A WITH MACRON
      case 0x102:    ch = 'A'; break; // LATIN CAPITAL LETTER A WITH BREVE
      case 0x103:    ch = 'a'; break; // LATIN SMALL LETTER A WITH BREVE
      case 0x104:    ch = 'A'; break; // LATIN CAPITAL LETTER A WITH OGONEK
      case 0x105:    ch = 'a'; break; // LATIN SMALL LETTER A WITH OGONEK
      case 0x106:    ch = 'C'; break; // LATIN CAPITAL LETTER C WITH ACUTE
      case 0x107:    ch = 'c'; break; // LATIN SMALL LETTER C WITH ACUTE
      case 0x108:    ch = 'C'; break; // LATIN CAPITAL LETTER C WITH CIRCUMFLEX
      case 0x109:    ch = 'c'; break; // LATIN SMALL LETTER C WITH CIRCUMFLEX
      case 0x10A:    ch = 'C'; break; // LATIN CAPITAL LETTER C WITH DOT ABOVE
      case 0x10B:    ch = 'c'; break; // LATIN SMALL LETTER C WITH DOT ABOVE
      case 0x10C:    ch = 'C'; break; // LATIN CAPITAL LETTER C WITH CARON
      case 0x10D:    ch = 'c'; break; // LATIN SMALL LETTER C WITH CARON
      case 0x10E:    ch = 'D'; break; // LATIN CAPITAL LETTER D WITH CARON
      case 0x10F:    ch = 'd'; break; // LATIN SMALL LETTER D WITH CARON
      case 0x110:    ch = 'D'; break; // LATIN CAPITAL LETTER D WITH STROKE
      case 0x111:    ch = 'd'; break; // LATIN SMALL LETTER D WITH STROKE

      case 0x112:    ch = 'E'; break; // LATIN CAPITAL LETTER E WITH MACRON
      case 0x113:    ch = 'e'; break; // LATIN SMALL LETTER E WITH MACRON
      case 0x114:    ch = 'E'; break; // LATIN CAPITAL LETTER E WITH BREVE
      case 0x115:    ch = 'e'; break; // LATIN SMALL LETTER E WITH BREVE
      case 0x116:    ch = 'E'; break; // LATIN CAPITAL LETTER E WITH DOT ABOVE
      case 0x117:    ch = 'e'; break; // LATIN SMALL LETTER E WITH DOT ABOVE
      case 0x118:    ch = 'E'; break; // LATIN CAPITAL LETTER E WITH OGONEK
      case 0x119:    ch = 'e'; break; // LATIN SMALL LETTER E WITH OGONEK
      case 0x11A:    ch = 'E'; break; // LATIN CAPITAL LETTER E WITH CARON
      case 0x11B:    ch = 'e'; break; // LATIN SMALL LETTER E WITH CARON
      case 0x11C:    ch = 'G'; break; // LATIN CAPITAL LETTER G WITH CIRCUMFLEX
      case 0x11D:    ch = 'g'; break; // LATIN SMALL LETTER G WITH CIRCUMFLEX
      case 0x11E:    ch = 'G'; break; // LATIN CAPITAL LETTER G WITH BREVE
      case 0x11F:    ch = 'g'; break; // LATIN SMALL LETTER G WITH BREVE
      case 0x120:    ch = 'G'; break; // LATIN CAPITAL LETTER G WITH DOT ABOVE
      case 0x121:    ch = 'g'; break; // LATIN SMALL LETTER G WITH DOT ABOVE
      case 0x122:    ch = 'G'; break; // LATIN CAPITAL LETTER G WITH CEDILLA
      case 0x123:    ch = 'g'; break; // LATIN SMALL LETTER G WITH CEDILLA
      case 0x124:    ch = 'H'; break; // LATIN CAPITAL LETTER H WITH CIRCUMFLEX
      case 0x125:    ch = 'h'; break; // LATIN SMALL LETTER H WITH CIRCUMFLEX
      case 0x126:    ch = 'H'; break; // LATIN CAPITAL LETTER H WITH STROKE
      case 0x127:    ch = 'h'; break; // LATIN SMALL LETTER H WITH STROKE
      case 0x128:    ch = 'I'; break; // LATIN CAPITAL LETTER I WITH TILDE
      case 0x129:    ch = 'i'; break; // LATIN SMALL LETTER I WITH TILDE
      case 0x12A:    ch = 'I'; break; // LATIN CAPITAL LETTER I WITH MACRON
      case 0x12B:    ch = 'i'; break; // LATIN SMALL LETTER I WITH MACRON
      case 0x12C:    ch = 'I'; break; // LATIN CAPITAL LETTER I WITH BREVE
      case 0x12D:    ch = 'i'; break; // LATIN SMALL LETTER I WITH BREVE
      case 0x12E:    ch = 'I'; break; // LATIN CAPITAL LETTER I WITH OGONEK
      case 0x12F:    ch = 'i'; break; // LATIN SMALL LETTER I WITH OGONEK
      case 0x130:    ch = 'I'; break; // LATIN CAPITAL LETTER I WITH DOT ABOVE
      case 0x131:    ch = 'i'; break; // LATIN SMALL LETTER DOTLESS I
      case 0x132:    ch = 'J'; break; // LATIN CAPITAL LIGATURE IJ
      case 0x133:    ch = 'j'; break; // LATIN SMALL LIGATURE IJ
      case 0x134:    ch = 'J'; break; // LATIN CAPITAL LETTER J WITH CIRCUMFLEX
      case 0x135:    ch = 'j'; break; // LATIN SMALL LETTER J WITH CIRCUMFLEX
      case 0x136:    ch = 'K'; break; // LATIN CAPITAL LETTER K WITH CEDILLA
      case 0x137:    ch = 'k'; break; // LATIN SMALL LETTER K WITH CEDILLA
      case 0x138:    ch = 'k'; break; // LATIN SMALL LETTER KRA
      case 0x139:    ch = 'L'; break; // LATIN CAPITAL LETTER L WITH ACUTE
      case 0x13A:    ch = 'l'; break; // LATIN SMALL LETTER L WITH ACUTE
      case 0x13B:    ch = 'L'; break; // LATIN CAPITAL LETTER L WITH CEDILLA
      case 0x13C:    ch = 'l'; break; // LATIN SMALL LETTER L WITH CEDILLA
      case 0x13D:    ch = 'L'; break; // LATIN CAPITAL LETTER L WITH CARON
      case 0x13E:    ch = 'l'; break; // LATIN SMALL LETTER L WITH CARON
      case 0x13F:    ch = 'L'; break; // LATIN CAPITAL LETTER L WITH MIDDLE DOT
      case 0x140:    ch = 'l'; break; // LATIN SMALL LETTER L WITH MIDDLE DOT
      case 0x141:    ch = 'L'; break; // LATIN CAPITAL LETTER L WITH STROKE
      case 0x142:    ch = 'l'; break; // LATIN SMALL LETTER L WITH STROKE
      case 0x143:    ch = 'N'; break; // LATIN CAPITAL LETTER N WITH ACUTE
      case 0x144:    ch = 'n'; break; // LATIN SMALL LETTER N WITH ACUTE
      case 0x145:    ch = 'N'; break; // LATIN CAPITAL LETTER N WITH CEDILLA
      case 0x146:    ch = 'n'; break; // LATIN SMALL LETTER N WITH CEDILLA
      case 0x147:    ch = 'N'; break; // LATIN CAPITAL LETTER N WITH CARON
      case 0x148:    ch = 'n'; break; // LATIN SMALL LETTER N WITH CARON
      case 0x149:    ch = 'n'; break; // LATIN SMALL LETTER N PRECEDED BY APOSTROPHE
      case 0x14A:    ch = 'E'; break; // LATIN CAPITAL LETTER ENG
      case 0x14B:    ch = 'e'; break; // LATIN SMALL LETTER ENG
      case 0x14C:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH MACRON
      case 0x14D:    ch = 'o'; break; // LATIN SMALL LETTER O WITH MACRON
      case 0x14E:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH BREVE
      case 0x14F:    ch = 'o'; break; // LATIN SMALL LETTER O WITH BREVE
      case 0x150:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH DOUBLE ACUTE
      case 0x151:    ch = 'o'; break; // LATIN SMALL LETTER O WITH DOUBLE ACUTE
      case 0x152:    ch = 'E'; break; // LATIN CAPITAL LIGATURE OE
      case 0x153:    ch = 'e'; break; // LATIN SMALL LIGATURE OE
      case 0x154:    ch = 'R'; break; // LATIN CAPITAL LETTER R WITH ACUTE
      case 0x155:    ch = 'r'; break; // LATIN SMALL LETTER R WITH ACUTE
      case 0x156:    ch = 'R'; break; // LATIN CAPITAL LETTER R WITH CEDILLA
      case 0x157:    ch = 'r'; break; // LATIN SMALL LETTER R WITH CEDILLA
      case 0x158:    ch = 'R'; break; // LATIN CAPITAL LETTER R WITH CARON
      case 0x159:    ch = 'r'; break; // LATIN SMALL LETTER R WITH CARON
      case 0x15A:    ch = 'S'; break; // LATIN CAPITAL LETTER S WITH ACUTE
      case 0x15B:    ch = 's'; break; // LATIN SMALL LETTER S WITH ACUTE
      case 0x15C:    ch = 'S'; break; // LATIN CAPITAL LETTER S WITH CIRCUMFLEX
      case 0x15D:    ch = 's'; break; // LATIN SMALL LETTER S WITH CIRCUMFLEX
      case 0x15E:    ch = 'S'; break; // LATIN CAPITAL LETTER S WITH CEDILLA
      case 0x15F:    ch = 's'; break; // LATIN SMALL LETTER S WITH CEDILLA
      case 0x160:    ch = 'S'; break; // LATIN CAPITAL LETTER S WITH CARON
      case 0x161:    ch = 's'; break; // LATIN SMALL LETTER S WITH CARON
      case 0x162:    ch = 'T'; break; // LATIN CAPITAL LETTER T WITH CEDILLA
      case 0x163:    ch = 't'; break; // LATIN SMALL LETTER T WITH CEDILLA
      case 0x164:    ch = 'T'; break; // LATIN CAPITAL LETTER T WITH CARON
      case 0x165:    ch = 't'; break; // LATIN SMALL LETTER T WITH CARON
      case 0x166:    ch = 'T'; break; // LATIN CAPITAL LETTER T WITH STROKE
      case 0x167:    ch = 't'; break; // LATIN SMALL LETTER T WITH STROKE
      case 0x168:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH TILDE
      case 0x169:    ch = 'u'; break; // LATIN SMALL LETTER U WITH TILDE
      case 0x16A:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH MACRON

      case 0x16B:    ch = 'u'; break; // LATIN SMALL LETTER U WITH MACRON
      case 0x16C:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH BREVE
      case 0x16D:    ch = 'u'; break; // LATIN SMALL LETTER U WITH BREVE
      case 0x16E:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH RING ABOVE
      case 0x16F:    ch = 'u'; break; // LATIN SMALL LETTER U WITH RING ABOVE
      case 0x170:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH DOUBLE ACUTE
      case 0x171:    ch = 'u'; break; // LATIN SMALL LETTER U WITH DOUBLE ACUTE
      case 0x172:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH OGONEK
      case 0x173:    ch = 'u'; break; // LATIN SMALL LETTER U WITH OGONEK
      case 0x174:    ch = 'W'; break; // LATIN CAPITAL LETTER W WITH CIRCUMFLEX
      case 0x175:    ch = 'w'; break; // LATIN SMALL LETTER W WITH CIRCUMFLEX
      case 0x176:    ch = 'Y'; break; // LATIN CAPITAL LETTER Y WITH CIRCUMFLEX
      case 0x177:    ch = 'y'; break; // LATIN SMALL LETTER Y WITH CIRCUMFLEX
      case 0x178:    ch = 'Y'; break; // LATIN CAPITAL LETTER Y WITH DIAERESIS
      case 0x179:    ch = 'Z'; break; // LATIN CAPITAL LETTER Z WITH ACUTE
      case 0x17A:    ch = 'z'; break; // LATIN SMALL LETTER Z WITH ACUTE
      case 0x17B:    ch = 'Z'; break; // LATIN CAPITAL LETTER Z WITH DOT ABOVE
      case 0x17C:    ch = 'z'; break; // LATIN SMALL LETTER Z WITH DOT ABOVE
      case 0x17D:    ch = 'Z'; break; // LATIN CAPITAL LETTER Z WITH CARON
      case 0x17E:    ch = 'z'; break; // LATIN SMALL LETTER Z WITH CARON

      case 0x17F:    ch = 'S'; break; // LATIN SMALL LETTER LONG S
      case 0x180:    ch = 'b'; break; // LATIN SMALL LETTER B WITH STROKE
      case 0x181:    ch = 'B'; break; // LATIN CAPITAL LETTER B WITH HOOK
      case 0x182:    ch = 'B'; break; // LATIN CAPITAL LETTER B WITH TOPBAR
      case 0x183:    ch = 'B'; break; // LATIN SMALL LETTER B WITH TOPBAR
      case 0x184:    ch = '6'; break; // LATIN CAPITAL LETTER TONE SIX
      case 0x185:    ch = '6'; break; // LATIN SMALL LETTER TONE SIX
      case 0x186:    ch = 'O'; break; // LATIN CAPITAL LETTER OPEN O
      case 0x187:    ch = 'C'; break; // LATIN CAPITAL LETTER C WITH HOOK
      case 0x188:    ch = 'c'; break; // LATIN SMALL LETTER C WITH HOOK
      case 0x189:    ch = 'D'; break; // LATIN CAPITAL LETTER AFRICAN D
      case 0x18A:    ch = 'D'; break; // LATIN CAPITAL LETTER D WITH HOOK
      case 0x18B:    ch = 'D'; break; // LATIN CAPITAL LETTER D WITH TOPBAR
      case 0x18C:    ch = 'd'; break; // LATIN SMALL LETTER D WITH TOPBAR
      case 0x18D:    ch = 'd'; break; // LATIN SMALL LETTER TURNED DELTA
      case 0x18E:    ch = 'E'; break; // LATIN CAPITAL LETTER REVERSED E
      case 0x18F:    ch = 'a'; break; // LATIN CAPITAL LETTER SCHWA
      case 0x190:    ch = 'E'; break; // LATIN CAPITAL LETTER OPEN E
      case 0x191:    ch = 'F'; break; // LATIN CAPITAL LETTER F WITH HOOK
      case 0x192:    ch = 'f'; break; // LATIN SMALL LETTER F WITH HOOK
      case 0x193:    ch = 'G'; break; // LATIN CAPITAL LETTER G WITH HOOK
      case 0x194:    ch = 'g'; break; // LATIN CAPITAL LETTER GAMMA
      case 0x195:    ch = 'h'; break; // LATIN SMALL LETTER HV
      case 0x196:    ch = 'I'; break; // LATIN CAPITAL LETTER IOTA
      case 0x197:    ch = 'I'; break; // LATIN CAPITAL LETTER I WITH STROKE
      case 0x198:    ch = 'I'; break; // LATIN CAPITAL LETTER K WITH HOOK
      case 0x199:    ch = 'k'; break; // LATIN SMALL LETTER K WITH HOOK
      case 0x19A:    ch = 'l'; break; // LATIN SMALL LETTER L WITH BAR
      case 0x19B:    ch = 'l'; break; // LATIN SMALL LETTER LAMBDA WITH STROKE
      case 0x19C:    ch = 'M'; break; // LATIN CAPITAL LETTER TURNED M
      case 0x19D:    ch = 'N'; break; // LATIN CAPITAL LETTER N WITH LEFT HOOK
      case 0x19E:    ch = 'n'; break; // LATIN SMALL LETTER N WITH LONG RIGHT LEG
      case 0x19F:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH MIDDLE TILDE
      case 0x1A0:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH HORN
      case 0x1A1:    ch = 'o'; break; // LATIN SMALL LETTER O WITH HORN
      case 0x1A2:    ch = 'O'; break; // LATIN CAPITAL LETTER OI
      case 0x1A3:    ch = 'o'; break; // LATIN SMALL LETTER OI
      case 0x1A4:    ch = 'P'; break; // LATIN CAPITAL LETTER P WITH HOOK
      case 0x1A5:    ch = 'p'; break; // LATIN SMALL LETTER P WITH HOOK
      case 0x1A6:    ch = 'Y'; break; // LATIN LETTER YR
      case 0x1A7:    ch = '2'; break; // LATIN CAPITAL LETTER TONE TWO
      case 0x1A8:    ch = '2'; break; // LATIN SMALL LETTER TONE TWO
      case 0x1A9:    ch = 'S'; break; // LATIN CAPITAL LETTER ESH
      case 0x1AA:    ch = 'S'; break; // LATIN LETTER REVERSED ESH LOOP
      case 0x1AB:    ch = 't'; break; // LATIN SMALL LETTER T WITH PALATAL HOOK
      case 0x1AC:    ch = 'T'; break; // LATIN CAPITAL LETTER T WITH HOOK
      case 0x1AD:    ch = 't'; break; // LATIN SMALL LETTER T WITH HOOK
      case 0x1AE:    ch = 'T'; break; // LATIN CAPITAL LETTER T WITH RETROFLEX HOOK
      case 0x1AF:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH HORN
      case 0x1B0:    ch = 'u'; break; // LATIN SMALL LETTER U WITH HORN
      case 0x1B1:    ch = 'U'; break; // LATIN CAPITAL LETTER UPSILON
      case 0x1B2:    ch = 'V'; break; // LATIN CAPITAL LETTER V WITH HOOK
      case 0x1B3:    ch = 'Y'; break; // LATIN CAPITAL LETTER Y WITH HOOK
      case 0x1B4:    ch = 'y'; break; // LATIN SMALL LETTER Y WITH HOOK
      case 0x1B5:    ch = 'Z'; break; // LATIN CAPITAL LETTER Z WITH STROKE
      case 0x1B6:    ch = 'z'; break; // LATIN SMALL LETTER Z WITH STROKE
      case 0x1B7:    ch = 'E'; break; // LATIN CAPITAL LETTER EZH
      case 0x1B8:    ch = 'E'; break; // LATIN CAPITAL LETTER EZH REVERSED
      case 0x1B9:    ch = 'e'; break; // LATIN SMALL LETTER EZH REVERSED
      case 0x1BA:    ch = 'e'; break; // LATIN SMALL LETTER EZH WITH TAIL
      case 0x1BB:    ch = '2'; break; // LATIN LETTER TWO WITH STROKE
      case 0x1BC:    ch = '5'; break; // LATIN CAPITAL LETTER TONE FIVE
      case 0x1BD:    ch = '5'; break; // LATIN SMALL LETTER TONE FIVE
      case 0x1BE:    ch = '!'; break; // LATIN LETTER INVERTED GLOTTAL STOP WITH STROKE
      case 0x1BF:    ch = 'W'; break; // LATIN LETTER WYNN
      case 0x1C0:    ch = 'K'; break; // LATIN LETTER DENTAL CLICK
      case 0x1C1:    ch = 'K'; break; // LATIN LETTER LATERAL CLICK
      case 0x1C2:    ch = 'K'; break; // LATIN LETTER ALVEOLAR CLICK
      case 0x1C3:    ch = 'K'; break; // LATIN LETTER RETROFLEX CLICK
      case 0x1C4:    ch = 'Z'; break; // LATIN CAPITAL LETTER DZ WITH CARON
      case 0x1C5:    ch = 'D'; break; // LATIN CAPITAL LETTER D WITH SMALL LETTER Z WITH CARON
      case 0x1C6:    ch = 'z'; break; // LATIN SMALL LETTER DZ WITH CARON
      case 0x1C7:    ch = 'J'; break; // LATIN CAPITAL LETTER LJ
      case 0x1C8:    ch = 'L'; break; // LATIN CAPITAL LETTER L WITH SMALL LETTER J
      case 0x1C9:    ch = 'l'; break; // LATIN SMALL LETTER LJ
      case 0x1CA:    ch = 'N'; break; // LATIN CAPITAL LETTER NJ
      case 0x1CB:    ch = 'N'; break; // LATIN CAPITAL LETTER N WITH SMALL LETTER J
      case 0x1CC:    ch = 'N'; break; // LATIN SMALL LETTER NJ
      case 0x1CD:    ch = 'A'; break; // LATIN CAPITAL LETTER A WITH CARON
      case 0x1CE:    ch = 'a'; break; // LATIN SMALL LETTER A WITH CARON
      case 0x1CF:    ch = 'I'; break; // LATIN CAPITAL LETTER I WITH CARON
      case 0x1D0:    ch = 'i'; break; // LATIN SMALL LETTER I WITH CARON
      case 0x1D1:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH CARON
      case 0x1D2:    ch = 'o'; break; // LATIN SMALL LETTER O WITH CARON
      case 0x1D3:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH CARON
      case 0x1D4:    ch = 'u'; break; // LATIN SMALL LETTER U WITH CARON
      case 0x1D5:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH DIAERESIS AND MACRON
      case 0x1D6:    ch = 'u'; break; // LATIN SMALL LETTER U WITH DIAERESIS AND MACRON
      case 0x1D7:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH DIAERESIS AND ACUTE
      case 0x1D8:    ch = 'u'; break; // LATIN SMALL LETTER U WITH DIAERESIS AND ACUTE
      case 0x1D9:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH DIAERESIS AND CARON
      case 0x1DA:    ch = 'u'; break; // LATIN SMALL LETTER U WITH DIAERESIS AND CARON
      case 0x1DB:    ch = 'U'; break; // LATIN CAPITAL LETTER U WITH DIAERESIS AND GRAVE
      case 0x1DC:    ch = 'u'; break; // LATIN SMALL LETTER U WITH DIAERESIS AND GRAVE
      case 0x1DD:    ch = 'u'; break; // LATIN SMALL LETTER TURNED E
      case 0x1DE:    ch = 'A'; break; // LATIN CAPITAL LETTER A WITH DIAERESIS AND MACRON
      case 0x1DF:    ch = 'a'; break; // LATIN SMALL LETTER A WITH DIAERESIS AND MACRON
      case 0x1E0:    ch = 'A'; break; // LATIN CAPITAL LETTER A WITH DOT ABOVE AND MACRON
      case 0x1E1:    ch = 'a'; break; // LATIN SMALL LETTER A WITH DOT ABOVE AND MACRON
      case 0x1E2:    ch = 'E'; break; // LATIN CAPITAL LETTER AE WITH MACRON
      case 0x1E3:    ch = 'e'; break; // LATIN SMALL LETTER AE WITH MACRON
      case 0x1E4:    ch = 'G'; break; // LATIN CAPITAL LETTER G WITH STROKE
      case 0x1E5:    ch = 'g'; break; // LATIN SMALL LETTER G WITH STROKE
      case 0x1E6:    ch = 'G'; break; // LATIN CAPITAL LETTER G WITH CARON
      case 0x1E7:    ch = 'g'; break; // LATIN SMALL LETTER G WITH CARON
      case 0x1E8:    ch = 'K'; break; // LATIN CAPITAL LETTER K WITH CARON
      case 0x1E9:    ch = 'k'; break; // LATIN SMALL LETTER K WITH CARON
      case 0x1EA:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH OGONEK
      case 0x1EB:    ch = 'o'; break; // LATIN SMALL LETTER O WITH OGONEK
      case 0x1EC:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH OGONEK AND MACRON
      case 0x1ED:    ch = 'o'; break; // LATIN SMALL LETTER O WITH OGONEK AND MACRON
      case 0x1EE:    ch = 'Z'; break; // LATIN CAPITAL LETTER EZH WITH CARON
      case 0x1EF:    ch = 'z'; break; // LATIN SMALL LETTER EZH WITH CARON
      case 0x1F0:    ch = 'J'; break; // LATIN SMALL LETTER J WITH CARON
      case 0x1F1:    ch = 'Z'; break; // LATIN CAPITAL LETTER DZ
      case 0x1F2:    ch = 'Z'; break; // LATIN CAPITAL LETTER D WITH SMALL LETTER Z
      case 0x1F3:    ch = 'z'; break; // LATIN SMALL LETTER DZ
      case 0x1F4:    ch = 'G'; break; // LATIN CAPITAL LETTER G WITH ACUTE
      case 0x1F5:    ch = 'g'; break; // LATIN SMALL LETTER G WITH ACUTE
      case 0x1F6:    ch = 'R'; break; // LATIN CAPITAL LETTER HWAIR
      case 0x1F7:    ch = 'W'; break; // LATIN CAPITAL LETTER WYNN
      case 0x1F8:    ch = 'N'; break; // LATIN CAPITAL LETTER N WITH GRAVE
      case 0x1F9:    ch = 'N'; break; // LATIN SMALL LETTER N WITH GRAVE
      case 0x1FA:    ch = 'A'; break; // LATIN CAPITAL LETTER A WITH RING ABOVE AND ACUTE
      case 0x1FB:    ch = 'a'; break; // LATIN SMALL LETTER A WITH RING ABOVE AND ACUTE
      case 0x1FC:    ch = 'E'; break; // LATIN CAPITAL LETTER AE WITH ACUTE
      case 0x1FD:    ch = 'e'; break; // LATIN SMALL LETTER AE WITH ACUTE
      case 0x1FE:    ch = 'O'; break; // LATIN CAPITAL LETTER O WITH STROKE AND ACUTE
      case 0x1FF:    ch = 'o'; break; // LATIN SMALL LETTER O WITH STROKE AND ACUTE

      case 0x200:    ch = 'A'; // LATIN CAPITAL LETTER A WITH DOUBLE GRAVE
      case 0x201:    ch = 'a'; // LATIN SMALL LETTER A WITH DOUBLE GRAVE
      case 0x202:    ch = 'A'; // LATIN CAPITAL LETTER A WITH INVERTED BREVE
      case 0x203:    ch = 'a'; // LATIN SMALL LETTER A WITH INVERTED BREVE
      case 0x204:    ch = 'E'; // LATIN CAPITAL LETTER E WITH DOUBLE GRAVE
      case 0x205:    ch = 'e'; // LATIN SMALL LETTER E WITH DOUBLE GRAVE
      case 0x206:    ch = 'E'; // LATIN CAPITAL LETTER E WITH INVERTED BREVE
      case 0x207:    ch = 'e'; // LATIN SMALL LETTER E WITH INVERTED BREVE
      case 0x208:    ch = 'I'; // LATIN CAPITAL LETTER I WITH DOUBLE GRAVE
      case 0x209:    ch = 'i'; // LATIN SMALL LETTER I WITH DOUBLE GRAVE
      case 0x20A:    ch = 'I'; // LATIN CAPITAL LETTER I WITH INVERTED BREVE
      case 0x20B:    ch = 'i'; // LATIN SMALL LETTER I WITH INVERTED BREVE
      case 0x20C:    ch = 'O'; // LATIN CAPITAL LETTER O WITH DOUBLE GRAVE
      case 0x20D:    ch = 'o'; // LATIN SMALL LETTER O WITH DOUBLE GRAVE
      case 0x20E:    ch = 'O'; // LATIN CAPITAL LETTER O WITH INVERTED BREVE
      case 0x20F:    ch = 'o'; // LATIN SMALL LETTER O WITH INVERTED BREVE
      case 0x210:    ch = 'R'; // LATIN CAPITAL LETTER R WITH DOUBLE GRAVE
      case 0x211:    ch = 'r'; // LATIN SMALL LETTER R WITH DOUBLE GRAVE
      case 0x212:    ch = 'R'; // LATIN CAPITAL LETTER R WITH INVERTED BREVE
      case 0x213:    ch = 'r'; // LATIN SMALL LETTER R WITH INVERTED BREVE
      case 0x214:    ch = 'U'; // LATIN CAPITAL LETTER U WITH DOUBLE GRAVE
      case 0x215:    ch = 'u'; // LATIN SMALL LETTER U WITH DOUBLE GRAVE
      case 0x216:    ch = 'U'; // LATIN CAPITAL LETTER U WITH INVERTED BREVE
      case 0x217:    ch = 'u'; // LATIN SMALL LETTER U WITH INVERTED BREVE
      case 0x218:    ch = 'S'; // LATIN CAPITAL LETTER S WITH COMMA BELOW
      case 0x219:    ch = 's'; // LATIN SMALL LETTER S WITH COMMA BELOW
      case 0x21A:    ch = 'T'; // LATIN CAPITAL LETTER T WITH COMMA BELOW
      case 0x21B:    ch = 't'; // LATIN SMALL LETTER T WITH COMMA BELOW
      case 0x21C:    ch = 'Y'; // LATIN CAPITAL LETTER YOGH
      case 0x21D:    ch = 'y'; // LATIN SMALL LETTER YOGH
      case 0x21E:    ch = 'H'; // LATIN CAPITAL LETTER H WITH CARON
      case 0x21F:    ch = 'h'; // LATIN SMALL LETTER H WITH CARON
      case 0x220:    ch = 'N'; // LATIN CAPITAL LETTER N WITH LONG RIGHT LEG
      case 0x221:    ch = 'D'; // LATIN SMALL LETTER D WITH CURL
      case 0x222:    ch = 'U'; // LATIN CAPITAL LETTER OU
      case 0x223:    ch = 'U'; // LATIN SMALL LETTER OU
      case 0x224:    ch = 'Z'; // LATIN CAPITAL LETTER Z WITH HOOK
      case 0x225:    ch = 'Z'; // LATIN SMALL LETTER Z WITH HOOK
      case 0x226:    ch = 'A'; // LATIN CAPITAL LETTER A WITH DOT ABOVE
      case 0x227:    ch = 'a'; // LATIN SMALL LETTER A WITH DOT ABOVE
      case 0x228:    ch = 'E'; // LATIN CAPITAL LETTER E WITH CEDILLA
      case 0x229:    ch = 'e'; // LATIN SMALL LETTER E WITH CEDILLA
      case 0x22A:    ch = 'O'; // LATIN CAPITAL LETTER O WITH DIAERESIS AND MACRON
      case 0x22B:    ch = 'o'; // LATIN SMALL LETTER O WITH DIAERESIS AND MACRON
      case 0x22C:    ch = 'O'; // LATIN CAPITAL LETTER O WITH TILDE AND MACRON
      case 0x22D:    ch = 'o'; // LATIN SMALL LETTER O WITH TILDE AND MACRON
      case 0x22E:    ch = 'O'; // LATIN CAPITAL LETTER O WITH DOT ABOVE
      case 0x22F:    ch = 'o'; // LATIN SMALL LETTER O WITH DOT ABOVE
      case 0x230:    ch = 'O'; // LATIN CAPITAL LETTER O WITH DOT ABOVE AND MACRON
      case 0x231:    ch = 'O'; // LATIN SMALL LETTER O WITH DOT ABOVE AND MACRON
      case 0x232:    ch = 'Y'; // LATIN CAPITAL LETTER Y WITH MACRON
      case 0x233:    ch = 'Y'; // LATIN SMALL LETTER Y WITH MACRON
      case 0x234:    ch = 'L'; // LATIN SMALL LETTER L WITH CURL
      case 0x235:    ch = 'N'; // LATIN SMALL LETTER N WITH CURL
      case 0x236:    ch = 'T'; // LATIN SMALL LETTER T WITH CURL
      case 0x237:    ch = 'J'; // LATIN SMALL LETTER DOTLESS J
      case 0x238:    ch = 'b'; // LATIN SMALL LETTER DB DIGRAPH
      case 0x239:    ch = 'p'; // LATIN SMALL LETTER QP DIGRAPH
      case 0x23A:    ch = 'A'; // LATIN CAPITAL LETTER A WITH STROKE
      case 0x23B:    ch = 'C'; // LATIN CAPITAL LETTER C WITH STROKE
      case 0x23C:    ch = 'c'; // LATIN SMALL LETTER C WITH STROKE
      case 0x23D:    ch = 'L'; // LATIN CAPITAL LETTER L WITH BAR
      case 0x23E:    ch = 'T'; // LATIN CAPITAL LETTER T WITH DIAGONAL STROKE
      case 0x23F:    ch = 'S'; // LATIN SMALL LETTER S WITH SWASH TAIL
      case 0x240:    ch = 'z'; // LATIN SMALL LETTER Z WITH SWASH TAIL
      case 0x241:    ch = '!'; // LATIN CAPITAL LETTER GLOTTAL STOP
      case 0x242:    ch = '!'; // LATIN SMALL LETTER GLOTTAL STOP
      case 0x243:    ch = 'B'; // LATIN CAPITAL LETTER B WITH STROKE
      case 0x244:    ch = 'U'; // LATIN CAPITAL LETTER U BAR
      case 0x245:    ch = 'V'; // LATIN CAPITAL LETTER TURNED V
      case 0x246:    ch = 'E'; // LATIN CAPITAL LETTER E WITH STROKE
      case 0x247:    ch = 'E'; // LATIN SMALL LETTER E WITH STROKE
      case 0x248:    ch = 'J'; // LATIN CAPITAL LETTER J WITH STROKE
      case 0x249:    ch = 'j'; // LATIN SMALL LETTER J WITH STROKE
      case 0x24A:    ch = 'Q'; // LATIN CAPITAL LETTER SMALL Q WITH HOOK TAIL
      case 0x24B:    ch = 'q'; // LATIN SMALL LETTER Q WITH HOOK TAIL
      case 0x24C:    ch = 'R'; // LATIN CAPITAL LETTER R WITH STROKE
      case 0x24D:    ch = 'r'; // LATIN SMALL LETTER R WITH STROKE
      case 0x24E:    ch = 'Y'; // LATIN CAPITAL LETTER Y WITH STROKE
      case 0x24F:    ch = 'y'; // LATIN SMALL LETTER Y WITH STROKE
      case 0x250:    ch = 'a'; // LATIN SMALL LETTER TURNED A
      case 0x251:    ch = '@'; // LATIN SMALL LETTER ALPHA
      case 0x252:    ch = '@'; // LATIN SMALL LETTER TURNED ALPHA
      case 0x253:    ch = 'b'; // LATIN SMALL LETTER B WITH HOOK
      case 0x254:    ch = 'o'; // LATIN SMALL LETTER OPEN O
      case 0x255:    ch = 'c'; // LATIN SMALL LETTER C WITH CURL
      case 0x256:    ch = 'd'; // LATIN SMALL LETTER D WITH TAIL
      case 0x257:    ch = 'e'; // LATIN SMALL LETTER D WITH HOOK
      case 0x258:    ch = 'w'; // LATIN SMALL LETTER REVERSED E
      case 0x259:    ch = 'w'; // LATIN SMALL LETTER SCHWA
      case 0x25A:    ch = 'w'; // LATIN SMALL LETTER SCHWA WITH HOOK
      case 0x25B:    ch = 'e'; // LATIN SMALL LETTER OPEN E
      case 0x25C:    ch = 'e'; // LATIN SMALL LETTER REVERSED OPEN E
      case 0x25D:    ch = 'e'; // LATIN SMALL LETTER REVERSED OPEN E WITH HOOK
      case 0x25E:    ch = 'e'; // LATIN SMALL LETTER CLOSED REVERSED OPEN E
      case 0x25F:    ch = 'j'; // LATIN SMALL LETTER DOTLESS J WITH STROKE
      case 0x260:    ch = 'g'; // LATIN SMALL LETTER G WITH HOOK
      case 0x261:    ch = 'g'; // LATIN SMALL LETTER SCRIPT G
      case 0x262:    ch = 'g'; // LATIN LETTER SMALL CAPITAL G
      case 0x263:    ch = 'r'; // LATIN SMALL LETTER GAMMA
      case 0x264:    ch = 'h'; // LATIN SMALL LETTER RAMS HORN
      case 0x265:    ch = 'h'; // LATIN SMALL LETTER TURNED H
      case 0x266:    ch = 'h'; // LATIN SMALL LETTER H WITH HOOK
      case 0x267:    ch = 'h'; // LATIN SMALL LETTER HENG WITH HOOK
      case 0x268:    ch = 'i'; // LATIN SMALL LETTER I WITH STROKE
      case 0x269:    ch = 'i'; // LATIN SMALL LETTER IOTA
      case 0x26A:    ch = 'i'; // LATIN LETTER SMALL CAPITAL I
      case 0x26B:    ch = 'l'; // LATIN SMALL LETTER L WITH MIDDLE TILDE
      case 0x26C:    ch = 'l'; // LATIN SMALL LETTER L WITH BELT
      case 0x26D:    ch = 'l'; // LATIN SMALL LETTER L WITH RETROFLEX HOOK
      case 0x26E:    ch = 'l'; // LATIN SMALL LETTER LEZH
      case 0x26F:    ch = 'm'; // LATIN SMALL LETTER TURNED M
      case 0x270:    ch = 'm'; // LATIN SMALL LETTER TURNED M WITH LONG LEG
      case 0x271:    ch = 'm'; // LATIN SMALL LETTER M WITH HOOK
      case 0x272:    ch = 'n'; // LATIN SMALL LETTER N WITH LEFT HOOK
      case 0x273:    ch = 'n'; // LATIN SMALL LETTER N WITH RETROFLEX HOOK
      case 0x274:    ch = 'N'; // LATIN LETTER SMALL CAPITAL N
      case 0x275:    ch = 'o'; // LATIN SMALL LETTER BARRED O
      case 0x276:    ch = '1'; // LATIN LETTER SMALL CAPITAL OE
      case 0x277:    ch = 'o'; // LATIN SMALL LETTER CLOSED OMEGA
      case 0x278:    ch = 'p'; // LATIN SMALL LETTER PHI
      case 0x279:    ch = 'r'; // LATIN SMALL LETTER TURNED R
      case 0x27A:    ch = 'r'; // LATIN SMALL LETTER TURNED R WITH LONG LEG
      case 0x27B:    ch = 'r'; // LATIN SMALL LETTER TURNED R WITH HOOK
      case 0x27C:    ch = 'r'; // LATIN SMALL LETTER R WITH LONG LEG
      case 0x27D:    ch = 'r'; // LATIN SMALL LETTER R WITH TAIL
      case 0x27E:    ch = 'r'; // LATIN SMALL LETTER R WITH FISHHOOK
      case 0x27F:    ch = 'r'; // LATIN SMALL LETTER REVERSED R WITH FISHHOOK
      case 0x280:    ch = 'r'; // LATIN LETTER SMALL CAPITAL R
      case 0x281:    ch = 'r'; // LATIN LETTER SMALL CAPITAL INVERTED R
      case 0x282:    ch = 's'; // LATIN SMALL LETTER S WITH HOOK
      case 0x283:    ch = 's'; // LATIN SMALL LETTER ESH
      case 0x284:    ch = 'j'; // LATIN SMALL LETTER DOTLESS J WITH STROKE AND HOOK
      case 0x285:    ch = 's'; // LATIN SMALL LETTER SQUAT REVERSED ESH
      case 0x286:    ch = 's'; // LATIN SMALL LETTER ESH WITH CURL
      case 0x287:    ch = 't'; // LATIN SMALL LETTER TURNED T
      case 0x288:    ch = 't'; // LATIN SMALL LETTER T WITH RETROFLEX HOOK
      case 0x289:    ch = 'u'; // LATIN SMALL LETTER U BAR
      case 0x28A:    ch = 'v'; // LATIN SMALL LETTER UPSILON
      case 0x28B:    ch = 'v'; // LATIN SMALL LETTER V WITH HOOK
      case 0x28C:    ch = 'v'; // LATIN SMALL LETTER TURNED V
      case 0x28D:    ch = 'w'; // LATIN SMALL LETTER TURNED W
      case 0x28E:    ch = 'y'; // LATIN SMALL LETTER TURNED Y
      case 0x28F:    ch = 'y'; // LATIN LETTER SMALL CAPITAL Y
      case 0x290:    ch = 'z'; // LATIN SMALL LETTER Z WITH RETROFLEX HOOK
      case 0x291:    ch = 'z'; // LATIN SMALL LETTER Z WITH CURL
      case 0x292:    ch = 'z'; // LATIN SMALL LETTER EZH
      case 0x293:    ch = 'z'; // LATIN SMALL LETTER EZH WITH CURL
      case 0x294:    ch = '!'; // LATIN LETTER GLOTTAL STOP
      case 0x295:    ch = '!'; // LATIN LETTER PHARYNGEAL VOICED FRICATIVE
      case 0x296:    ch = '!'; // LATIN LETTER INVERTED GLOTTAL STOP
      case 0x297:    ch = 'c'; // LATIN LETTER STRETCHED C
      case 0x298:    ch = 'k'; // LATIN LETTER BILABIAL CLICK
      case 0x299:    ch = 'b'; // LATIN LETTER SMALL CAPITAL B
      case 0x29A:    ch = 'e'; // LATIN SMALL LETTER CLOSED OPEN E
      case 0x29B:    ch = 'g'; // LATIN LETTER SMALL CAPITAL G WITH HOOK
      case 0x29C:    ch = 'h'; // LATIN LETTER SMALL CAPITAL H
      case 0x29D:    ch = 'j'; // LATIN SMALL LETTER J WITH CROSSED-TAIL
      case 0x29E:    ch = 'k'; // LATIN SMALL LETTER TURNED K
      case 0x29F:    ch = 'l'; // LATIN LETTER SMALL CAPITAL L
      case 0x2A0:    ch = 'q'; // LATIN SMALL LETTER Q WITH HOOK
      case 0x2A1:    ch = '!'; // LATIN LETTER GLOTTAL STOP WITH STROKE
      case 0x2A2:    ch = '!'; // LATIN LETTER REVERSED GLOTTAL STOP WITH STROKE
      case 0x2A3:    ch = 'z'; // LATIN SMALL LETTER DZ DIGRAPH
      case 0x2A4:    ch = 'z'; // LATIN SMALL LETTER DEZH DIGRAPH
      case 0x2A5:    ch = 'z'; // LATIN SMALL LETTER DZ DIGRAPH WITH CURL
      case 0x2A6:    ch = 'z'; // LATIN SMALL LETTER TS DIGRAPH
      case 0x2A7:    ch = 'z'; // LATIN SMALL LETTER TESH DIGRAPH
      case 0x2A8:    ch = 'z'; // LATIN SMALL LETTER TC DIGRAPH WITH CURL
      case 0x2A9:    ch = 'g'; // LATIN SMALL LETTER FENG DIGRAPH
      case 0x2AA:    ch = 's'; // LATIN SMALL LETTER LS DIGRAPH
      case 0x2AB:    ch = 'z'; // LATIN SMALL LETTER LZ DIGRAPH
      case 0x2AC:    ch = 'p'; // LATIN LETTER BILABIAL PERCUSSIVE
      case 0x2AD:    ch = 'p'; // LATIN LETTER BIDENTAL PERCUSSIVE
      case 0x2AE:    ch = 'h'; // LATIN SMALL LETTER TURNED H WITH FISHHOOK
      case 0x2AF:    ch = 'h'; // LATIN SMALL LETTER TURNED H WITH FISHHOOK AND TAIL
      case 0x2B0:    ch = 'h'; // MODIFIER LETTER SMALL H
      case 0x2B1:    ch = 'h'; // MODIFIER LETTER SMALL H WITH HOOK
      case 0x2B2:    ch = 'j'; // MODIFIER LETTER SMALL J
      case 0x2B3:    ch = 'r'; // MODIFIER LETTER SMALL R
      case 0x2B4:    ch = 'r'; // MODIFIER LETTER SMALL TURNED R
      case 0x2B5:    ch = 'r'; // MODIFIER LETTER SMALL TURNED R WITH HOOK
      case 0x2B6:    ch = 'r'; // MODIFIER LETTER SMALL CAPITAL INVERTED R
      case 0x2B7:    ch = 'w'; // MODIFIER LETTER SMALL W
      case 0x2B8:    ch = 'y'; // MODIFIER LETTER SMALL Y
      case 0x2B9:    ch = '\''; // MODIFIER LETTER PRIME
      case 0x2BA:    ch = '"'; // MODIFIER LETTER DOUBLE PRIME
      case 0x2BB:    ch = ','; // MODIFIER LETTER TURNED COMMA
      case 0x2BC:    ch = '\''; // MODIFIER LETTER APOSTROPHE
      case 0x2BD:    ch = ','; // MODIFIER LETTER REVERSED COMMA
      case 0x2BE:    ch = ','; // MODIFIER LETTER RIGHT HALF RING
      case 0x2BF:    ch = 'r'; // MODIFIER LETTER LEFT HALF RING
      case 0x2C0:    ch = '!'; // MODIFIER LETTER GLOTTAL STOP
      case 0x2C1:    ch = '!'; // MODIFIER LETTER REVERSED GLOTTAL STOP
      case 0x2C2:    ch = '<'; // MODIFIER LETTER LEFT ARROWHEAD
      case 0x2C3:    ch = '>'; // MODIFIER LETTER RIGHT ARROWHEAD
      case 0x2C4:    ch = '^'; // MODIFIER LETTER UP ARROWHEAD
      case 0x2C5:    ch = 'V'; // MODIFIER LETTER DOWN ARROWHEAD
      case 0x2C6:    ch = '~'; // MODIFIER LETTER CIRCUMFLEX ACCENT
      case 0x2C7:    ch = '_'; // CARON
      case 0x2C8:    ch = '|'; // MODIFIER LETTER VERTICAL LINE
      case 0x2C9:    ch = '-'; // MODIFIER LETTER MACRON
      case 0x2CA:    ch = '`'; // MODIFIER LETTER ACUTE ACCENT
      case 0x2CB:    ch = '\''; // MODIFIER LETTER GRAVE ACCENT
      case 0x2CC:    ch = '_'; // MODIFIER LETTER LOW VERTICAL LINE
      case 0x2CD:    ch = '-'; // MODIFIER LETTER LOW MACRON
      case 0x2CE:    ch = '!'; // MODIFIER LETTER LOW GRAVE ACCENT
      case 0x2CF:    ch = '!'; // MODIFIER LETTER LOW ACUTE ACCENT
      case 0x2D0:    ch = 'v'; // MODIFIER LETTER TRIANGULAR COLON
      case 0x2D1:    ch = '>'; // MODIFIER LETTER HALF TRIANGULAR COLON
      case 0x2D2:    ch = '>'; // MODIFIER LETTER CENTRED RIGHT HALF RING
      case 0x2D3:    ch = '<'; // MODIFIER LETTER CENTRED LEFT HALF RING
      case 0x2D4:    ch = '^'; // MODIFIER LETTER UP TACK
      case 0x2D5:    ch = 'V'; // MODIFIER LETTER DOWN TACK
      case 0x2D6:    ch = '+'; // MODIFIER LETTER PLUS SIGN
      case 0x2D7:    ch = '-'; // MODIFIER LETTER MINUS SIGN
      case 0x2D8:    ch = '!'; // BREVE
      case 0x2D9:    ch = '.'; // DOT ABOVE
      case 0x2DA:    ch = 'o'; // RING ABOVE
      case 0x2DB:    ch = 'O'; // OGONEK
      case 0x2DC:    ch = '~'; // SMALL TILDE
      case 0x2DD:    ch = '~'; // DOUBLE ACUTE ACCENT
      case 0x2DE:    ch = 'r'; // MODIFIER LETTER RHOTIC HOOK
      case 0x2DF:    ch = 'x'; // MODIFIER LETTER CROSS ACCENT
      case 0x2E0:    ch = 'x'; // MODIFIER LETTER SMALL GAMMA
      case 0x2E1:    ch = 'l'; // MODIFIER LETTER SMALL L
      case 0x2E2:    ch = 's'; // MODIFIER LETTER SMALL S
      case 0x2E3:    ch = 'x'; // MODIFIER LETTER SMALL X
      case 0x2E4:    ch = '!'; // MODIFIER LETTER SMALL REVERSED GLOTTAL STOP
      case 0x2E5:    ch = '!'; // MODIFIER LETTER EXTRA-HIGH TONE BAR
      case 0x2E6:    ch = '!'; // MODIFIER LETTER HIGH TONE BAR
      case 0x2E7:    ch = '!'; // MODIFIER LETTER MID TONE BAR
      case 0x2E8:    ch = '!'; // MODIFIER LETTER LOW TONE BAR
      case 0x2E9:    ch = '|'; // MODIFIER LETTER EXTRA-LOW TONE BAR
      case 0x2EA:    ch = 'Y'; // MODIFIER LETTER YIN DEPARTING TONE MARK
      case 0x2EB:    ch = 'y'; // MODIFIER LETTER YANG DEPARTING TONE MARK
      case 0x2EC:    ch = 'V'; // MODIFIER LETTER VOICING
      case 0x2ED:    ch = 'u'; // MODIFIER LETTER UNASPIRATED
      case 0x2EE:    ch = '"'; // MODIFIER LETTER DOUBLE APOSTROPHE
      case 0x2EF:    ch = 'V'; // MODIFIER LETTER LOW DOWN ARROWHEAD
      case 0x2F0:    ch = '^'; // MODIFIER LETTER LOW UP ARROWHEAD
      case 0x2F1:    ch = '<'; // MODIFIER LETTER LOW LEFT ARROWHEAD
      case 0x2F2:    ch = '>'; // MODIFIER LETTER LOW RIGHT ARROWHEAD
      case 0x2F3:    ch = 'o'; // MODIFIER LETTER LOW RING
      case 0x2F4:    ch = '`'; // MODIFIER LETTER MIDDLE GRAVE ACCENT
      case 0x2F5:    ch = '`'; // MODIFIER LETTER MIDDLE DOUBLE GRAVE ACCENT
      case 0x2F6:    ch = '`'; // MODIFIER LETTER MIDDLE DOUBLE ACUTE ACCENT
      case 0x2F7:    ch = '~'; // MODIFIER LETTER LOW TILDE
      case 0x2F8:    ch = ':'; // MODIFIER LETTER RAISED COLON
      case 0x2F9:    ch = ':'; // MODIFIER LETTER BEGIN HIGH TONE
      case 0x2FA:    ch = ':'; // MODIFIER LETTER END HIGH TONE
      case 0x2FB:    ch = ':'; // MODIFIER LETTER BEGIN LOW TONE
      case 0x2FC:    ch = ':'; // MODIFIER LETTER END LOW TONE
      case 0x2FD:    ch = '_'; // MODIFIER LETTER SHELF
      case 0x2FE:    ch = '_'; // MODIFIER LETTER OPEN SHELF
      case 0x2FF:    ch = '_'; // MODIFIER LETTER LOW LEFT ARROW

      case 0x2010:   ch = '-'; break;  //  e2 80 90  HYPHEN
      case 0x2011:   ch = '-'; break;  //  e2 80 91  NON-BREAKING HYPHEN
      case 0x2012:   ch = '-'; break;  //  e2 80 92  FIGURE DASH
      case 0x2013:   ch = '-'; break;  //  e2 80 93  EN DASH
      case 0x2014:   ch = '-'; break;  //  e2 80 94  EM DASH
      case 0x2015:   ch = '_'; break;  //  e2 80 95  HORIZONTAL BAR
      case 0x2016:   ch = '|'; break;  //  e2 80 96  DOUBLE VERTICAL LINE
      case 0x2017:   ch = '|'; break;  //  e2 80 97  DOUBLE LOW LINE
      case 0x2018:   ch = '\''; break;  //  e2 80 98  LEFT SINGLE QUOTATION MARK
      case 0x2019:   ch = '\''; break;  //  e2 80 99  RIGHT SINGLE QUOTATION MARK
      case 0x201A:   ch = '"'; break;  //  e2 80 9a  SINGLE LOW-9 QUOTATION MARK
      case 0x201B:   ch = '"'; break;  //  e2 80 9b  SINGLE HIGH-REVERSED-9 QUOTATION MARK
      case 0x201C:   ch = '"'; break;  //  e2 80 9c  LEFT DOUBLE QUOTATION MARK
      case 0x201D:   ch = '"'; break;  //  e2 80 9d  RIGHT DOUBLE QUOTATION MARK
      case 0x201E:   ch = '"'; break;  //  e2 80 9e  DOUBLE LOW-9 QUOTATION MARK
      case 0x201F:   ch = '"'; break;  //  e2 80 9f  DOUBLE HIGH-REVERSED-9 QUOTATION MARK
      case 0x2020:   ch = '\\'; break;  //  e2 80 a0  DAGGER
      case 0x2021:   ch = '\\'; break;  //  e2 80 a1  DOUBLE DAGGER
      case 0x2022:   ch = '>'; break;  //  e2 80 a2  BULLET
      case 0x2023:   ch = 'v'; break;  //  e2 80 a3  TRIANGULAR BULLET
      case 0x2024:   ch = '.'; break;  //  e2 80 a4  ONE DOT LEADER
      case 0x2025:   ch = '.'; break;  //  e2 80 a5  TWO DOT LEADER
      case 0x2026:   ch = '.'; break;  //  e2 80 a6  HORIZONTAL ELLIPSIS
      case 0x2027:   ch = '.'; break;  //  e2 80 a7  HYPHENATION POINT


      // ARG!

   }

   return ch;
} // conflate

encoding fileEncoding(char const *filename)
{
    int fh = open(filename, O_RDONLY 
#ifdef _MSC_VER
                            | O_BINARY
#endif
                 );


    if(fh < 0)
    {
        return utfunk_enc;
    }

    char buffer[32];

    int bytesRead = read(fh, buffer, 32);

    close(fh);  // we won't keep reading it here

    if(bytesRead < 1 )
    {
        return utfunk_enc;   // no data return, could be empty file or some kind of error    
    }

    if(bytesRead == 1)
    {
       return utf7_enc;  // basically, it ain't utf
    }

    char *scan = buffer;

    return streamEncoding(scan, scan + bytesRead);


}

} // namespace utf
} // namespace cxxtls
